import abc
from typing import (Mapping, MutableMapping, List, Union, Iterable,
                    Iterator, TypeVar, Generic, Tuple, Dict, Optional,
                    overload)


class istr(str):
    pass


upstr = istr

_S = Union[str, istr]

_T = TypeVar('_T')

_T_co = TypeVar('_T_co', covariant=True)

_D = TypeVar('_D')


class MultiMapping(Mapping[_S, _T_co]):
    @overload
    @abc.abstractmethod
    def getall(self, key: _S) -> List[_T_co]: ...

    @overload
    @abc.abstractmethod
    def getall(self, key: _S, default: _D) -> Union[List[_T_co], _D]: ...

    @overload
    @abc.abstractmethod
    def getone(self, key: _S) -> _T_co: ...

    @overload
    @abc.abstractmethod
    def getone(self, key: _S, default: _D) -> Union[_T_co, _D]: ...


_Arg = Union[Mapping[_S, _T],
             Dict[_S, _T],
             MultiMapping[_T],
             Iterable[Tuple[_S, _T]]]


class MutableMultiMapping(MultiMapping[_T],
                          MutableMapping[_S, _T],
                          Generic[_T]):

    @abc.abstractmethod
    def add(self, key: _S, value: _T) -> None: ...

    @abc.abstractmethod
    def extend(self, arg: _Arg[_T] = ..., **kwargs: _T) -> None: ...

    @overload
    @abc.abstractmethod
    def popone(self, key: _S) -> _T: ...

    @overload
    @abc.abstractmethod
    def popone(self, key: _S, default: _D) -> Union[_T, _D]: ...

    @overload
    @abc.abstractmethod
    def popall(self, key: _S) -> List[_T]: ...

    @overload
    @abc.abstractmethod
    def popall(self, key: _S, default: _D) -> Union[List[_T], _D]: ...


class MultiDict(MutableMultiMapping[_T], Generic[_T]):
    def __init__(self, arg: _Arg[_T] = ..., **kwargs: _T) -> None: ...

    def copy(self) -> MultiDict[_T]: ...

    def __getitem__(self, k: _S) -> _T: ...

    def __setitem__(self, k: _S, v: _T) -> None: ...

    def __delitem__(self, v: _S) -> None: ...

    def __iter__(self) -> Iterator[_S]: ...

    def __len__(self) -> int: ...

    @overload
    def getall(self, key: _S) -> List[_T]: ...
    @overload
    def getall(self, key: _S, default: _D) -> Union[List[_T], _D]: ...

    @overload
    def getone(self, key: _S) -> _T: ...
    @overload
    def getone(self, key: _S, default: _D) -> Union[_T, _D]: ...

    def add(self, key: _S, value: _T) -> None: ...

    def extend(self, arg: _Arg[_T] = ..., **kwargs: _T) -> None: ...

    @overload
    def popone(self, key: _S) -> _T: ...
    @overload
    def popone(self, key: _S, default: _D) -> Union[_T, _D]: ...

    @overload
    def popall(self, key: _S) -> List[_T]: ...
    @overload
    def popall(self, key: _S, default: _D) -> Union[List[_T], _D]: ...


class CIMultiDict(MutableMultiMapping[_T], Generic[_T]):
    def __init__(self, arg: _Arg[_T] = ..., **kwargs: _T) -> None: ...

    def copy(self) -> MultiDict[_T]: ...

    def __getitem__(self, k: _S) -> _T: ...

    def __setitem__(self, k: _S, v: _T) -> None: ...

    def __delitem__(self, v: _S) -> None: ...

    def __iter__(self) -> Iterator[_S]: ...

    def __len__(self) -> int: ...

    @overload
    def getall(self, key: _S) -> List[_T]: ...
    @overload
    def getall(self, key: _S, default: _D) -> Union[List[_T], _D]: ...

    @overload
    def getone(self, key: _S) -> _T: ...
    @overload
    def getone(self, key: _S, default: _D) -> Union[_T, _D]: ...

    def add(self, key: _S, value: _T) -> None: ...

    def extend(self, arg: _Arg[_T] = ..., **kwargs: _T) -> None: ...

    @overload
    def popone(self, key: _S) -> _T: ...
    @overload
    def popone(self, key: _S, default: _D) -> Union[_T, _D]: ...

    @overload
    def popall(self, key: _S) -> List[_T]: ...
    @overload
    def popall(self, key: _S, default: _D) -> Union[List[_T], _D]: ...


class MultiDictProxy(MultiMapping[_T], Generic[_T]):
    def __init__(self, arg: Union[MultiMapping[_T],
                                  MutableMultiMapping[_T]]) -> None: ...

    def copy(self) -> MultiDict[_T]: ...

    def __getitem__(self, k: _S) -> _T: ...

    def __iter__(self) -> Iterator[_S]: ...

    def __len__(self) -> int: ...

    @overload
    def getall(self, key: _S) -> List[_T]: ...
    @overload
    def getall(self, key: _S, default: _D) -> Union[List[_T], _D]: ...

    @overload
    def getone(self, key: _S) -> _T: ...
    @overload
    def getone(self, key: _S, default: _D) -> Union[_T, _D]: ...


class CIMultiDictProxy(MultiMapping[_T], Generic[_T]):
    def __init__(self, arg: Union[MultiMapping[_T],
                                  MutableMultiMapping[_T]]) -> None: ...

    def __getitem__(self, k: _S) -> _T: ...

    def __iter__(self) -> Iterator[_S]: ...

    def __len__(self) -> int: ...

    @overload
    def getall(self, key: _S) -> List[_T]: ...
    @overload
    def getall(self, key: _S, default: _D) -> Union[List[_T], _D]: ...

    @overload
    def getone(self, key: _S) -> _T: ...
    @overload
    def getone(self, key: _S, default: _D) -> Union[_T, _D]: ...

    def copy(self) -> CIMultiDict[_T]: ...


def getversion(md: Union[MultiDict[_T],
                         CIMultiDict[_T],
                         MultiDictProxy[_T],
                         CIMultiDictProxy[_T]]) -> int: ...



from cpython.object cimport PyObject
from libc.stdint cimport uint64_t

cdef extern from "_pair_list.h":

    object pair_list_new()
    object ci_pair_list_new()

    int pair_list_len(object lst) except -1

    int pair_list_clear(object lst) except -1

    int _pair_list_add_with_hash(object lst,
                                 object identity, object key,
                                 object value, Py_hash_t hash) except -1
    int pair_list_add(object lst, object key, object value) except -1

    int _pair_list_next(object lst, Py_ssize_t *ppos,
                        PyObject* *pidentity,
                        PyObject* *pkey, PyObject* *pvalue,
                        Py_hash_t *hash) except -1

    int pair_list_next(object lst, Py_ssize_t *ppos,
                       PyObject* *pidentity,
                       PyObject* *pkey, PyObject* *pvalue) except -1

    int pair_list_contains(object lst, object key) except -1
    # todo: add key param to raise proper exception key
    object pair_list_get_one(object lst, object key)
    object pair_list_get_all(object lst, object key)

    int pair_list_del(object lst, object key) except -1

    object pair_list_set_default(object lst, object key, object value)

    object pair_list_pop_one(object lst, object key)
    object pair_list_pop_all(object lst, object key)
    object pair_list_pop_item(object lst)

    int pair_list_replace(object lst, object key, object value) except -1


    int pair_list_update(object list, object other) except -1
    int pair_list_update_from_seq(object list, object seq) except -1
    int pair_list_eq_to_mapping(object list, object other) except -1

    uint64_t pair_list_version(object lst)

    int pair_list_init(object istr_type) except -1



from cpython.object cimport PyObject

cdef extern from "_multidict_iter.h":

    object multidict_items_iter_new(object impl)
    object multidict_keys_iter_new(object impl)
    object multidict_values_iter_new(object impl)

    int multidict_iter_init() except -1



import pytest

from multidict._compat import USE_CYTHON

if USE_CYTHON:
    from multidict._multidict import (MultiDict, CIMultiDict)

from multidict._multidict_py import (MultiDict as PyMultiDict,  # noqa: E402
                                     CIMultiDict as PyCIMultiDict)


@pytest.fixture(
    params=(
        [
            MultiDict,
            CIMultiDict,
        ]
        if USE_CYTHON else
        []
    ) +
    [
        PyMultiDict,
        PyCIMultiDict
    ],
    ids=(
        [
            'MultiDict',
            'CIMultiDict',
        ]
        if USE_CYTHON else
        []
    ) +
    [
        'PyMultiDict',
        'PyCIMultiDict'
    ]
)
def cls(request):
    return request.param


@pytest.fixture
def md_cls(_multidict):
    return _multidict.MultiDict


@pytest.fixture
def ci_md_cls(_multidict):
    return _multidict.CIMultiDict


@pytest.fixture
def istr(_multidict):
    return _multidict.istr


def test_update_replace(cls):
    obj1 = cls([('a', 1), ('b', 2), ('a', 3), ('c', 10)])
    obj2 = cls([('a', 4), ('b', 5), ('a', 6)])
    obj1.update(obj2)
    expected = [('a', 4), ('b', 5), ('a', 6), ('c', 10)]
    assert list(obj1.items()) == expected


def test_update_append(cls):
    obj1 = cls([('a', 1), ('b', 2), ('a', 3), ('c', 10)])
    obj2 = cls([('a', 4), ('a', 5), ('a', 6)])
    obj1.update(obj2)
    expected = [('a', 4), ('b', 2), ('a', 5), ('c', 10), ('a', 6)]
    assert list(obj1.items()) == expected


def test_update_remove(cls):
    obj1 = cls([('a', 1), ('b', 2), ('a', 3), ('c', 10)])
    obj2 = cls([('a', 4)])
    obj1.update(obj2)
    expected = [('a', 4), ('b', 2), ('c', 10)]
    assert list(obj1.items()) == expected


def test_update_replace_seq(cls):
    obj1 = cls([('a', 1), ('b', 2), ('a', 3), ('c', 10)])
    obj2 = [('a', 4), ('b', 5), ('a', 6)]
    obj1.update(obj2)
    expected = [('a', 4), ('b', 5), ('a', 6), ('c', 10)]
    assert list(obj1.items()) == expected


def test_update_replace_seq2(cls):
    obj1 = cls([('a', 1), ('b', 2), ('a', 3), ('c', 10)])
    obj1.update([('a', 4)], b=5, a=6)
    expected = [('a', 4), ('b', 5), ('a', 6), ('c', 10)]
    assert list(obj1.items()) == expected


def test_update_append_seq(cls):
    obj1 = cls([('a', 1), ('b', 2), ('a', 3), ('c', 10)])
    obj2 = [('a', 4), ('a', 5), ('a', 6)]
    obj1.update(obj2)
    expected = [('a', 4), ('b', 2), ('a', 5), ('c', 10), ('a', 6)]
    assert list(obj1.items()) == expected


def test_update_remove_seq(cls):
    obj1 = cls([('a', 1), ('b', 2), ('a', 3), ('c', 10)])
    obj2 = [('a', 4)]
    obj1.update(obj2)
    expected = [('a', 4), ('b', 2), ('c', 10)]
    assert list(obj1.items()) == expected


def test_update_md(md_cls):
    d = md_cls()
    d.add('key', 'val1')
    d.add('key', 'val2')
    d.add('key2', 'val3')

    d.update(key='val')

    assert [('key', 'val'), ('key2', 'val3')] == list(d.items())


def test_update_istr_ci_md(ci_md_cls, istr):
    d = ci_md_cls()
    d.add(istr('KEY'), 'val1')
    d.add('key', 'val2')
    d.add('key2', 'val3')

    d.update({istr('key'): 'val'})

    assert [('Key', 'val'), ('key2', 'val3')] == list(d.items())


def test_update_ci_md(ci_md_cls):
    d = ci_md_cls()
    d.add('KEY', 'val1')
    d.add('key', 'val2')
    d.add('key2', 'val3')

    d.update(Key='val')

    assert [('Key', 'val'), ('key2', 'val3')] == list(d.items())



import copy

import pytest

from multidict._compat import USE_CYTHON

if USE_CYTHON:
    from multidict._multidict import (MultiDict, CIMultiDict,
                                      MultiDictProxy, CIMultiDictProxy)

from multidict._multidict_py import (MultiDict as PyMultiDict,  # noqa: E402
                                     CIMultiDict as PyCIMultiDict,
                                     MultiDictProxy as PyMultiDictProxy,
                                     CIMultiDictProxy as PyCIMultiDictProxy)


@pytest.fixture(
    params=(
        [
            MultiDict,
            CIMultiDict,
        ]
        if USE_CYTHON else
        []
    ) +
    [
        PyMultiDict,
        PyCIMultiDict
    ],
    ids=(
        [
            'MultiDict',
            'CIMultiDict',
        ]
        if USE_CYTHON else
        []
    ) +
    [
        'PyMultiDict',
        'PyCIMultiDict'
    ]
)
def cls(request):
    return request.param


@pytest.fixture(
    params=(
        [
            (MultiDictProxy, MultiDict),
            (CIMultiDictProxy, CIMultiDict),
        ]
        if USE_CYTHON else
        []
    ) +
    [
        (PyMultiDictProxy, PyMultiDict),
        (PyCIMultiDictProxy, PyCIMultiDict),
    ],
    ids=(
        [
            'MultiDictProxy',
            'CIMultiDictProxy',
        ]
        if USE_CYTHON else
        []
    ) +
    [
        'PyMultiDictProxy',
        'PyCIMultiDictProxy'
    ]
)
def proxy_classes(request):
    return request.param


def test_copy(cls):
    d = cls()
    d['foo'] = 6
    d2 = d.copy()
    d2['foo'] = 7
    assert d['foo'] == 6
    assert d2['foo'] == 7


def test_copy_proxy(proxy_classes):
    proxy_cls, dict_cls = proxy_classes
    d = dict_cls()
    d['foo'] = 6
    p = proxy_cls(d)
    d2 = p.copy()
    d2['foo'] = 7
    assert d['foo'] == 6
    assert p['foo'] == 6
    assert d2['foo'] == 7


def test_copy_std_copy(cls):
    d = cls()
    d['foo'] = 6
    d2 = copy.copy(d)
    d2['foo'] = 7
    assert d['foo'] == 6
    assert d2['foo'] == 7


def test_ci_multidict_clone(cls):
    d = cls(foo=6)
    d2 = cls(d)
    d2['foo'] = 7
    assert d['foo'] == 6
    assert d2['foo'] == 7



from cpython.object cimport PyObject

cdef extern from "_multidict_views.h":

    object multidict_itemsview_new(object impl)
    object multidict_keysview_new(object impl)
    object multidict_valuesview_new(object impl)

    int multidict_views_init() except -1



import pickle

from multidict._compat import USE_CYTHON

try:
    from multidict._multidict import MultiDict, CIMultiDict  # noqa
except ImportError:
    pass

from multidict._multidict_py import (MultiDict as PyMultiDict,  # noqa
                                     CIMultiDict as PyCIMultiDict)


def write(name, proto):
    cls = globals()[name]
    d = cls([('a', 1), ('a', 2)])
    with open('{}.pickle.{}'.format(name.lower(), proto), 'wb') as f:
        pickle.dump(d, f, proto)


def generate():
    if not USE_CYTHON:
        raise RuntimeError("Cython is required")
    for proto in range(pickle.HIGHEST_PROTOCOL + 1):
        for name in ('MultiDict', 'CIMultiDict',
                     'PyMultiDict', 'PyCIMultiDict'):
            write(name, proto)


if __name__ == '__main__':
    generate()



from collections import abc
from collections.abc import Set, Iterable


def _abc_itemsview_register(view_cls):
    abc.ItemsView.register(view_cls)


def _abc_keysview_register(view_cls):
    abc.KeysView.register(view_cls)


def _abc_valuesview_register(view_cls):
    abc.ValuesView.register(view_cls)


def _viewbaseset_richcmp(view, other, op):
    if op == 0:  # <
        if not isinstance(other, Set):
            return NotImplemented
        return len(view) < len(other) and view <= other
    elif op == 1:  # <=
        if not isinstance(other, Set):
            return NotImplemented
        if len(view) > len(other):
            return False
        for elem in view:
            if elem not in other:
                return False
        return True
    elif op == 2:  # ==
        if not isinstance(other, Set):
            return NotImplemented
        return len(view) == len(other) and view <= other
    elif op == 3:  # !=
        return not view == other
    elif op == 4:  # >
        if not isinstance(other, Set):
            return NotImplemented
        return len(view) > len(other) and view >= other
    elif op == 5:  # >=
        if not isinstance(other, Set):
            return NotImplemented
        if len(view) < len(other):
            return False
        for elem in other:
            if elem not in view:
                return False
        return True


def _viewbaseset_and(view, other):
    if not isinstance(other, Iterable):
        return NotImplemented
    if isinstance(view, Set):
        view = set(iter(view))
    if isinstance(other, Set):
        other = set(iter(other))
    if not isinstance(other, Set):
        other = set(iter(other))
    return view & other


def _viewbaseset_or(view, other):
    if not isinstance(other, Iterable):
        return NotImplemented
    if isinstance(view, Set):
        view = set(iter(view))
    if isinstance(other, Set):
        other = set(iter(other))
    if not isinstance(other, Set):
        other = set(iter(other))
    return view | other


def _viewbaseset_sub(view, other):
    if not isinstance(other, Iterable):
        return NotImplemented
    if isinstance(view, Set):
        view = set(iter(view))
    if isinstance(other, Set):
        other = set(iter(other))
    if not isinstance(other, Set):
        other = set(iter(other))
    return view - other


def _viewbaseset_xor(view, other):
    if not isinstance(other, Iterable):
        return NotImplemented
    if isinstance(view, Set):
        view = set(iter(view))
    if isinstance(other, Set):
        other = set(iter(other))
    if not isinstance(other, Set):
        other = set(iter(other))
    return view ^ other


def _itemsview_isdisjoint(view, other):
    'Return True if two sets have a null intersection.'
    for v in other:
        if v in view:
            return False
    return True


def _itemsview_repr(view):
    lst = []
    for k, v in view:
        lst.append("{!r}: {!r}".format(k, v))
    body = ', '.join(lst)
    return '{}({})'.format(view.__class__.__name__, body)


def _keysview_isdisjoint(view, other):
    'Return True if two sets have a null intersection.'
    for k in other:
        if k in view:
            return False
    return True


def _keysview_repr(view):
    lst = []
    for k in view:
        lst.append("{!r}".format(k))
    body = ', '.join(lst)
    return '{}({})'.format(view.__class__.__name__, body)


def _valuesview_repr(view):
    lst = []
    for v in view:
        lst.append("{!r}".format(v))
    body = ', '.join(lst)
    return '{}({})'.format(view.__class__.__name__, body)



from __future__ import absolute_import

import sys
from collections import abc
from collections.abc import Iterable, Set

from cpython.object cimport PyObject_Str, Py_NE, PyObject_RichCompare

from ._abc import MultiMapping, MutableMultiMapping
from ._istr import istr

from ._multidict_iter cimport *
from ._multidict_views cimport *
from ._pair_list cimport *

cdef object _marker = object()

upstr = istr  # for relaxing backward compatibility problems
cdef object _istr = istr

pair_list_init(istr)
# multidict_iter_init()
multidict_views_init()

def getversion(_Base md):
    return pair_list_version(md._impl)


cdef class _Base:

    cdef object _impl

    def impl(self):
        return self._impl

    def getall(self, key, default=_marker):
        """Return a list of all values matching the key."""
        try:
            return pair_list_get_all(self._impl, key)
        except KeyError:
            if default is not _marker:
                return default
            else:
                raise

    def getone(self, key, default=_marker):
        """Get first value matching the key."""
        return self._getone(key, default)

    cdef _getone(self, key, default):
        try:
            return pair_list_get_one(self._impl, key)
        except KeyError:
            if default is not _marker:
                return default
            else:
                raise

    # Mapping interface #

    def __getitem__(self, key):
        return self._getone(key, _marker)

    def get(self, key, default=None):
        """Get first value matching the key.

        The method is alias for .getone().
        """
        return self._getone(key, default)

    def __contains__(self, key):
        return self._contains(key)

    cdef _contains(self, key):
        return pair_list_contains(self._impl, key)

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return pair_list_len(self._impl)

    cpdef keys(self):
        """Return a new view of the dictionary's keys."""
        return multidict_keysview_new(self)

    def items(self):
        """Return a new view of the dictionary's items *(key, value) pairs)."""
        return multidict_itemsview_new(self)

    def values(self):
        """Return a new view of the dictionary's values."""
        return multidict_valuesview_new(self)

    def __repr__(self):
        lst = []
        for k, v in self.items():
            lst.append("'{}': {!r}".format(k, v))
        body = ', '.join(lst)
        return '<{}({})>'.format(self.__class__.__name__, body)

    def __eq__(self, arg):
        cdef Py_ssize_t pos1
        cdef PyObject *identity1
        cdef PyObject *value1
        cdef Py_hash_t h1

        cdef Py_ssize_t pos2
        cdef PyObject *identity2
        cdef PyObject *value2
        cdef Py_hash_t h2

        cdef _Base other

        if isinstance(arg, _Base):
            other = <_Base>arg
            if pair_list_len(self._impl) != pair_list_len(other._impl):
                return False
            pos1 = pos2 = 0
            while (_pair_list_next(self._impl, &pos1, &identity1,
                                   NULL, &value1, &h1) and
                   _pair_list_next(other._impl, &pos2, &identity2,
                                   NULL, &value2, &h2)):
                if h1 != h2:
                    return False
                if PyObject_RichCompare(<object>identity1, <object>identity2, Py_NE):
                    return False
                if PyObject_RichCompare(<object>value1, <object>value2, Py_NE):
                    return False
            return True
        elif isinstance(arg, abc.Mapping):
            return bool(pair_list_eq_to_mapping(self._impl, arg))
        else:
            return NotImplemented


cdef class MultiDictProxy(_Base):
    _proxy_classes = (MultiDict, MultiDictProxy)
    _base_class = MultiDict

    def __init__(self, arg):
        cdef _Base base
        if not isinstance(arg, self._proxy_classes):
            raise TypeError(
                'ctor requires {} instance'
                ', not {}'.format(
                    ' or '.join(self._proxy_classes),
                    type(arg)))

        base = arg
        self._impl = base._impl

    def __reduce__(self):
        raise TypeError("can't pickle {} objects"
                        .format(self.__class__.__name__))

    def copy(self):
        """Return a copy of itself."""
        return self._base_class(self)

MultiMapping.register(MultiDictProxy)


cdef class CIMultiDictProxy(MultiDictProxy):
    _proxy_classes = (CIMultiDict, CIMultiDictProxy)
    _base_class = CIMultiDict


MultiMapping.register(CIMultiDictProxy)


cdef str _str(key):
    typ = type(key)
    if typ is str:
        return <str>key
    if typ is _istr:
        return PyObject_Str(key)
    elif issubclass(typ, str):
        return str(key)
    else:
        raise TypeError("MultiDict keys should be either str "
                        "or subclasses of str")


cdef class MultiDict(_Base):
    """An ordered dictionary that can have multiple values for each key."""

    def __init__(self, *args, **kwargs):
        self._impl = pair_list_new()
        self._extend(args, kwargs, 'MultiDict', True)

    def __reduce__(self):
        return (
            self.__class__,
            (list(self.items()),)
        )

    cdef _extend(self, tuple args, dict kwargs, name, bint do_add):
        cdef object key
        cdef object value
        cdef object arg
        cdef object i

        if len(args) > 1:
            raise TypeError("{} takes at most 1 positional argument"
                            " ({} given)".format(name, len(args)))

        if args:
            arg = args[0]
            if isinstance(arg, _Base) and not kwargs:
                if do_add:
                    self._append_items((<_Base>arg)._impl)
                else:
                    self._update_items((<_Base>arg)._impl)
            else:
                if hasattr(arg, 'items'):
                    arg = arg.items()
                if kwargs:
                    arg = list(arg)
                    arg.extend(list(kwargs.items()))
                if do_add:
                    self._append_items_seq(arg, name)
                else:
                    pair_list_update_from_seq(self._impl, arg)
        else:
            arg = list(kwargs.items())
            if do_add:
                self._append_items_seq(arg, name)
            else:
                pair_list_update_from_seq(self._impl, arg)

    cdef object _update_items(self, object impl):
        pair_list_update(self._impl, impl)

    cdef object _append_items(self, object impl):
        cdef PyObject *key
        cdef PyObject *val
        cdef Py_ssize_t pos
        pos = 0
        while _pair_list_next(impl, &pos, NULL, &key, &val, NULL):
            self._add(<object>key, <object>val)

    cdef object _append_items_seq(self, object arg, object name):
        cdef object i
        cdef object key
        cdef object value
        for i in arg:
            if not len(i) == 2:
                raise TypeError(
                    "{} takes either dict or list of (key, value) "
                    "tuples".format(name))
            key = i[0]
            value = i[1]
            self._add(key, value)

    cdef _add(self, key, value):
        pair_list_add(self._impl, key, value);

    cdef _replace(self, key, value):
        pair_list_replace(self._impl, key, value)

    def add(self, key, value):
        """Add the key and value, not overwriting any previous value."""
        self._add(key, value)

    def copy(self):
        """Return a copy of itself."""
        ret = MultiDict()
        ret._extend((list(self.items()),), {}, 'copy', True)
        return ret

    def extend(self, *args, **kwargs):
        """Extend current MultiDict with more values.

        This method must be used instead of update.
        """
        self._extend(args, kwargs, "extend", True)

    def clear(self):
        """Remove all items from MultiDict"""
        pair_list_clear(self._impl)

    # MutableMapping interface #

    def __setitem__(self, key, value):
        self._replace(key, value)

    def __delitem__(self, key):
        pair_list_del(self._impl, key)

    def setdefault(self, key, default=None):
        """Return value for key, set value to default if key is not present."""
        return pair_list_set_default(self._impl, key, default)

    def popone(self, key, default=_marker):
        """Remove the last occurrence of key and return the corresponding
        value.

        If key is not found, default is returned if given, otherwise
        KeyError is raised.

        """
        try:
            return pair_list_pop_one(self._impl, key)
        except KeyError:
            if default is _marker:
                raise
            else:
                return default

    pop = popone

    def popall(self, key, default=_marker):
        """Remove all occurrences of key and return the list of corresponding
        values.

        If key is not found, default is returned if given, otherwise
        KeyError is raised.

        """
        try:
            return pair_list_pop_all(self._impl, key)
        except KeyError:
            if default is _marker:
                raise
            else:
                return default

    def popitem(self):
        """Remove and return an arbitrary (key, value) pair."""
        return pair_list_pop_item(self._impl)

    def update(self, *args, **kwargs):
        """Update the dictionary from *other*, overwriting existing keys."""
        self._extend(args, kwargs, "update", False)


MutableMultiMapping.register(MultiDict)


cdef class CIMultiDict(MultiDict):
    """An ordered dictionary that can have multiple values for each key."""

    def __init__(self, *args, **kwargs):
        self._impl = ci_pair_list_new()
        self._extend(args, kwargs, 'CIMultiDict', True)

    def __reduce__(self):
        return (
            self.__class__,
            (list(self.items()),),
        )

    def copy(self):
        """Return a copy of itself."""
        ret = CIMultiDict()
        ret._extend((list(self.items()),), {}, 'copy', True)
        return ret


MutableMultiMapping.register(CIMultiDict)




import pickle

import pytest

from multidict._compat import USE_CYTHON_EXTENSIONS


OPTIONAL_CYTHON = (() if USE_CYTHON_EXTENSIONS else
                   pytest.mark.skip(reason='No Cython extensions available'))


@pytest.fixture(scope='session', params=[
    pytest.param('multidict._multidict', marks=OPTIONAL_CYTHON),
    'multidict._multidict_py'
])
def _multidict(request):
    return pytest.importorskip(request.param)


def pytest_generate_tests(metafunc):
    if 'pickle_protocol' in metafunc.fixturenames:
        metafunc.parametrize("pickle_protocol",
                             list(range(pickle.HIGHEST_PROTOCOL + 1)),
                             scope='session')



import pytest

from multidict._compat import USE_CYTHON

if USE_CYTHON:
    from multidict._multidict import MultiDict, CIMultiDict, getversion

from multidict._multidict_py import (MultiDict as _MultiDict,  # noqa: E402
                                     CIMultiDict as _CIMultiDict,
                                     getversion as _getversion)


class VersionMixin:
    cls = NotImplemented

    def getver(self, md):
        raise NotImplementedError

    def test_getversion_bad_param(self):
        with pytest.raises(TypeError):
            self.getver(1)

    def test_ctor(self):
        m1 = self.cls()
        v1 = self.getver(m1)
        m2 = self.cls()
        v2 = self.getver(m2)
        assert v1 != v2

    def test_add(self):
        m = self.cls()
        v = self.getver(m)
        m.add('key', 'val')
        assert self.getver(m) > v

    def test_delitem(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        del m['key']
        assert self.getver(m) > v

    def test_delitem_not_found(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        with pytest.raises(KeyError):
            del m['notfound']
        assert self.getver(m) == v

    def test_setitem(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        m['key'] = 'val2'
        assert self.getver(m) > v

    def test_setitem_not_found(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        m['notfound'] = 'val2'
        assert self.getver(m) > v

    def test_clear(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        m.clear()
        assert self.getver(m) > v

    def test_setdefault(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        m.setdefault('key2', 'val2')
        assert self.getver(m) > v

    def test_popone(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        m.popone('key')
        assert self.getver(m) > v

    def test_popone_default(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        m.popone('key2', 'default')
        assert self.getver(m) == v

    def test_popone_key_error(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        with pytest.raises(KeyError):
            m.popone('key2')
        assert self.getver(m) == v

    def test_pop(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        m.pop('key')
        assert self.getver(m) > v

    def test_pop_default(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        m.pop('key2', 'default')
        assert self.getver(m) == v

    def test_pop_key_error(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        with pytest.raises(KeyError):
            m.pop('key2')
        assert self.getver(m) == v

    def test_popall(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        m.popall('key')
        assert self.getver(m) > v

    def test_popall_default(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        m.popall('key2', 'default')
        assert self.getver(m) == v

    def test_popall_key_error(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        with pytest.raises(KeyError):
            m.popall('key2')
        assert self.getver(m) == v

    def test_popitem(self):
        m = self.cls()
        m.add('key', 'val')
        v = self.getver(m)
        m.popitem()
        assert self.getver(m) > v

    def test_popitem_key_error(self):
        m = self.cls()
        v = self.getver(m)
        with pytest.raises(KeyError):
            m.popitem()
        assert self.getver(m) == v


if USE_CYTHON:
    class TestMultiDict(VersionMixin):

        cls = MultiDict

        def getver(self, md):
            return getversion(md)


if USE_CYTHON:
    class TestCIMultiDict(VersionMixin):

        cls = CIMultiDict

        def getver(self, md):
            return getversion(md)


class TestPyMultiDict(VersionMixin):

    cls = _MultiDict

    def getver(self, md):
        return _getversion(md)


class TestPyCIMultiDict(VersionMixin):

    cls = _CIMultiDict

    def getver(self, md):
        return _getversion(md)



"""Multidict implementation.

HTTP Headers and URL query string require specific data structure:
multidict. It behaves mostly like a dict but it can have
several values for the same key.
"""

from ._compat import USE_CYTHON_EXTENSIONS


__all__ = ('MultiMapping', 'MutableMultiMapping',
           'MultiDictProxy', 'CIMultiDictProxy',
           'MultiDict', 'CIMultiDict', 'upstr', 'istr')

__version__ = '4.5.2'


from ._abc import MultiMapping, MutableMultiMapping

try:
    if not USE_CYTHON_EXTENSIONS:
        raise ImportError
    from ._multidict import (MultiDictProxy,
                             CIMultiDictProxy,
                             MultiDict,
                             CIMultiDict,
                             upstr, istr)
except ImportError:  # pragma: no cover
    from ._multidict_py import (MultiDictProxy,
                                CIMultiDictProxy,
                                MultiDict,
                                CIMultiDict,
                                upstr, istr)



import operator
import sys
from functools import reduce

import pytest

import multidict


def chained_callable(module, callables):
    """
    Returns callable that will get and call all given objects in module in
    exact order. If `names` is a single object's name function will return
    object itself.

    Will treat `names` of type `str` as a list of single element.
    """
    callables = (callables,) if isinstance(callables, str) else callables
    _callable, *rest = (getattr(module, name) for name in callables)

    def chained_call(*args, **kwargs):
        return reduce(lambda res, c: c(res), rest, _callable(*args, **kwargs))

    return chained_call if len(rest) > 0 else _callable


@pytest.fixture(scope='function')
def cls(request, _multidict):
    return chained_callable(_multidict, request.param)


dict_cls = proxy_cls = cls


@pytest.mark.parametrize('cls', ['MultiDict', 'CIMultiDict'], indirect=True)
def test_exposed_names(cls):
    name = cls.__name__

    while name.startswith('_'):
        name = name[1:]

    assert name in multidict.__all__


@pytest.mark.parametrize('cls, key_cls', [
    ('MultiDict', str),
    (('MultiDict', 'MultiDictProxy'), str)
], indirect=['cls'])
def test__iter__types(cls, key_cls):
    d = cls([('key', 'one'), ('key2', 'two'), ('key', 3)])
    for i in d:
        assert type(i) is key_cls, (type(i), key_cls)


@pytest.mark.parametrize('dict_cls, proxy_cls', [
    ('MultiDict', 'MultiDictProxy'),
    ('CIMultiDict', 'CIMultiDictProxy'),
], indirect=True)
def test_proxy_copy(dict_cls, proxy_cls):
    d1 = dict_cls(key='value', a='b')
    p1 = proxy_cls(d1)

    d2 = p1.copy()
    assert d1 == d2
    assert d1 is not d2


class BaseMultiDictTest:

    def test_instantiate__empty(self, cls):
        d = cls()
        assert d == {}
        assert len(d) == 0
        assert list(d.keys()) == []
        assert list(d.values()) == []
        assert list(d.items()) == []

        assert cls() != list()
        with pytest.raises(TypeError, match=r'(2 given)'):
            cls(('key1', 'value1'), ('key2', 'value2'))

    @pytest.mark.parametrize('arg0', [
        [('key', 'value1')],
        {'key': 'value1'}
    ])
    def test_instantiate__from_arg0(self, cls, arg0):
        d = cls(arg0)

        assert d == {'key': 'value1'}
        assert len(d) == 1
        assert list(d.keys()) == ['key']
        assert list(d.values()) == ['value1']
        assert list(d.items()) == [('key', 'value1')]

    def test_instantiate__with_kwargs(self, cls):
        d = cls([('key', 'value1')], key2='value2')

        assert d == {'key': 'value1', 'key2': 'value2'}
        assert len(d) == 2
        assert sorted(d.keys()) == ['key', 'key2']
        assert sorted(d.values()) == ['value1', 'value2']
        assert sorted(d.items()) == [('key', 'value1'),
                                     ('key2', 'value2')]

    def test_instantiate__from_generator(self, cls):
        d = cls((str(i), i) for i in range(2))

        assert d == {'0': 0, '1': 1}
        assert len(d) == 2
        assert sorted(d.keys()) == ['0', '1']
        assert sorted(d.values()) == [0, 1]
        assert sorted(d.items()) == [('0', 0), ('1', 1)]

    def test_getone(self, cls):
        d = cls([('key', 'value1')], key='value2')

        assert d.getone('key') == 'value1'
        assert d.get('key') == 'value1'
        assert d['key'] == 'value1'

        with pytest.raises(KeyError, match='key2'):
            d['key2']
        with pytest.raises(KeyError, match='key2'):
            d.getone('key2')

        assert d.getone('key2', 'default') == 'default'

    def test__iter__(self, cls, ):
        d = cls([('key', 'one'), ('key2', 'two'), ('key', 3)])
        assert list(d) == ['key', 'key2', 'key']

    def test_keys__contains(self, cls):
        d = cls([('key', 'one'), ('key2', 'two'), ('key', 3)])

        assert list(d.keys()) == ['key', 'key2', 'key']

        assert 'key' in d.keys()
        assert 'key2' in d.keys()

        assert 'foo' not in d.keys()

    def test_values__contains(self, cls):
        d = cls([('key', 'one'), ('key', 'two'), ('key', 3)])

        assert list(d.values()) == ['one', 'two', 3]

        assert 'one' in d.values()
        assert 'two' in d.values()
        assert 3 in d.values()

        assert 'foo' not in d.values()

    def test_items__contains(self, cls):
        d = cls([('key', 'one'), ('key', 'two'), ('key', 3)])

        assert list(d.items()) == [('key', 'one'), ('key', 'two'), ('key', 3)]

        assert ('key', 'one') in d.items()
        assert ('key', 'two') in d.items()
        assert ('key', 3) in d.items()

        assert ('foo', 'bar') not in d.items()

    def test_cannot_create_from_unaccepted(self, cls):
        with pytest.raises(TypeError):
            cls([(1, 2, 3)])

    def test_keys_is_set_less(self, cls):
        d = cls([('key', 'value1')])

        assert d.keys() < {'key', 'key2'}

    def test_keys_is_set_less_equal(self, cls):
        d = cls([('key', 'value1')])

        assert d.keys() <= {'key'}

    def test_keys_is_set_equal(self, cls):
        d = cls([('key', 'value1')])

        assert d.keys() == {'key'}

    def test_keys_is_set_greater(self, cls):
        d = cls([('key', 'value1')])

        assert {'key', 'key2'} > d.keys()

    def test_keys_is_set_greater_equal(self, cls):
        d = cls([('key', 'value1')])

        assert {'key'} >= d.keys()

    def test_keys_is_set_not_equal(self, cls):
        d = cls([('key', 'value1')])

        assert d.keys() != {'key2'}

    def test_eq(self, cls):
        d = cls([('key', 'value1')])

        assert {'key': 'value1'} == d

    def test_eq2(self, cls):
        d1 = cls([('key', 'value1')])
        d2 = cls([('key2', 'value1')])

        assert d1 != d2

    def test_eq3(self, cls):
        d1 = cls([('key', 'value1')])
        d2 = cls()

        assert d1 != d2

    def test_eq_other_mapping_contains_more_keys(self, cls):
        d1 = cls(foo='bar')
        d2 = dict(foo='bar', bar='baz')

        assert d1 != d2

    def test_ne(self, cls):
        d = cls([('key', 'value1')])

        assert d != {'key': 'another_value'}

    def test_and(self, cls):
        d = cls([('key', 'value1')])

        assert {'key'} == d.keys() & {'key', 'key2'}

    def test_and2(self, cls):
        d = cls([('key', 'value1')])

        assert {'key'} == {'key', 'key2'} & d.keys()

    def test_or(self, cls):
        d = cls([('key', 'value1')])

        assert {'key', 'key2'} == d.keys() | {'key2'}

    def test_or2(self, cls):
        d = cls([('key', 'value1')])

        assert {'key', 'key2'} == {'key2'} | d.keys()

    def test_sub(self, cls):
        d = cls([('key', 'value1'), ('key2', 'value2')])

        assert {'key'} == d.keys() - {'key2'}

    def test_sub2(self, cls):
        d = cls([('key', 'value1'), ('key2', 'value2')])

        assert {'key3'} == {'key', 'key2', 'key3'} - d.keys()

    def test_xor(self, cls):
        d = cls([('key', 'value1'), ('key2', 'value2')])

        assert {'key', 'key3'} == d.keys() ^ {'key2', 'key3'}

    def test_xor2(self, cls):
        d = cls([('key', 'value1'), ('key2', 'value2')])

        assert {'key', 'key3'} == {'key2', 'key3'} ^ d.keys()

    @pytest.mark.parametrize('_set, expected', [
        ({'key2'}, True),
        ({'key'}, False)
    ])
    def test_isdisjoint(self, cls, _set, expected):
        d = cls([('key', 'value1')])

        assert d.keys().isdisjoint(_set) == expected

    def test_repr_issue_410(self, cls):
        d = cls()

        try:
            raise Exception
            pytest.fail("Should never happen")  # pragma: no cover
        except Exception as e:
            repr(d)

            assert sys.exc_info()[1] == e

    @pytest.mark.parametrize('op', [
        operator.or_, operator.and_, operator.sub, operator.xor
    ])
    @pytest.mark.parametrize('other', [{'other'}])
    def test_op_issue_410(self, cls, op, other):
        d = cls([('key', 'value')])

        try:
            raise Exception
            pytest.fail("Should never happen")  # pragma: no cover
        except Exception as e:
            op(d.keys(), other)

            assert sys.exc_info()[1] == e


class TestMultiDict(BaseMultiDictTest):

    @pytest.fixture(params=[
        'MultiDict',
        ('MultiDict', 'MultiDictProxy'),
    ])
    def cls(self, request, _multidict):
        return chained_callable(_multidict, request.param)

    def test__repr__(self, cls):
        d = cls()
        _cls = type(d)

        assert str(d) == '<%s()>' % _cls.__name__

        d = cls([('key', 'one'), ('key', 'two')])

        assert str(d) == "<%s('key': 'one', 'key': 'two')>" % _cls.__name__

    def test_getall(self, cls):
        d = cls([('key', 'value1')], key='value2')

        assert d != {'key': 'value1'}
        assert len(d) == 2

        assert d.getall('key') == ['value1', 'value2']

        with pytest.raises(KeyError, match='some_key'):
            d.getall('some_key')

        default = object()
        assert d.getall('some_key', default) is default

    def test_preserve_stable_ordering(self, cls):
        d = cls([('a', 1), ('b', '2'), ('a', 3)])
        s = '&'.join('{}={}'.format(k, v) for k, v in d.items())

        assert s == 'a=1&b=2&a=3'

    def test_get(self, cls):
        d = cls([('a', 1), ('a', 2)])
        assert d['a'] == 1

    def test_items__repr__(self, cls):
        d = cls([('key', 'value1')], key='value2')
        expected = "_ItemsView('key': 'value1', 'key': 'value2')"
        assert repr(d.items()) == expected

    def test_keys__repr__(self, cls):
        d = cls([('key', 'value1')], key='value2')
        assert repr(d.keys()) == "_KeysView('key', 'key')"

    def test_values__repr__(self, cls):
        d = cls([('key', 'value1')], key='value2')
        assert repr(d.values()) == "_ValuesView('value1', 'value2')"


class TestCIMultiDict(BaseMultiDictTest):

    @pytest.fixture(params=[
        'CIMultiDict',
        ('CIMultiDict', 'CIMultiDictProxy'),
    ])
    def cls(self, request, _multidict):
        return chained_callable(_multidict, request.param)

    def test_basics(self, cls):
        d = cls([('KEY', 'value1')], KEY='value2')

        assert d.getone('key') == 'value1'
        assert d.get('key') == 'value1'
        assert d.get('key2', 'val') == 'val'
        assert d['key'] == 'value1'
        assert 'key' in d

        with pytest.raises(KeyError, match='key2'):
            d['key2']
        with pytest.raises(KeyError, match='key2'):
            d.getone('key2')

    def test_getall(self, cls):
        d = cls([('KEY', 'value1')], KEY='value2')

        assert not d == {'KEY': 'value1'}
        assert len(d) == 2

        assert d.getall('key') == ['value1', 'value2']

        with pytest.raises(KeyError, match='some_key'):
            d.getall('some_key')

    def test_get(self, cls):
        d = cls([('A', 1), ('a', 2)])
        assert 1 == d['a']

    def test__repr__(self, cls):
        d = cls([('KEY', 'value1')], key='value2')
        _cls = type(d)

        expected = "<%s('KEY': 'value1', 'key': 'value2')>" % _cls.__name__
        assert str(d) == expected

    def test_items__repr__(self, cls):
        d = cls([('KEY', 'value1')], key='value2')
        expected = "_ItemsView('KEY': 'value1', 'key': 'value2')"
        assert repr(d.items()) == expected

    def test_keys__repr__(self, cls):
        d = cls([('KEY', 'value1')], key='value2')
        assert repr(d.keys()) == "_KeysView('KEY', 'key')"

    def test_values__repr__(self, cls):
        d = cls([('KEY', 'value1')], key='value2')
        assert repr(d.values()) == "_ValuesView('value1', 'value2')"



import gc
import sys

import pytest

from multidict._compat import USE_CYTHON

if USE_CYTHON:
    from multidict._multidict import istr

from multidict._multidict_py import istr as _istr  # noqa: E402


IMPLEMENTATION = getattr(sys, 'implementation')  # to suppress mypy error


class IStrMixin:

    cls = NotImplemented

    def test_ctor(self):
        s = self.cls()
        assert '' == s

    def test_ctor_str(self):
        s = self.cls('a')
        assert 'A' == s

    def test_ctor_str_uppercase(self):
        s = self.cls('A')
        assert 'A' == s

    def test_ctor_istr(self):
        s = self.cls('A')
        s2 = self.cls(s)
        assert 'A' == s
        assert s is s2

    def test_ctor_buffer(self):
        s = self.cls(b'a')
        assert "B'A'" == s

    def test_ctor_repr(self):
        s = self.cls(None)
        assert 'None' == s

    def test_title(self):
        s = self.cls('a')
        assert s is s.title()

    def test_str(self):
        s = self.cls('a')
        s1 = str(s)
        assert s1 == 'A'
        assert type(s1) is str

    def xtest_eq(self):
        s1 = 'Abc'
        s2 = self.cls(s1)
        assert s1 == s2
        assert s1.lower() == s2


class TestPyIStr(IStrMixin):
    cls = _istr

    @staticmethod
    def _create_strs():
        _istr('foobarbaz')
        istr2 = _istr()
        _istr(istr2)

    @pytest.mark.skipif(IMPLEMENTATION.name != 'cpython',
                        reason="PyPy has different GC implementation")
    def test_leak(self):
        gc.collect()
        cnt = len(gc.get_objects())
        for _ in range(10000):
            self._create_strs()

        gc.collect()
        cnt2 = len(gc.get_objects())
        assert abs(cnt - cnt2) < 10  # on PyPy these numbers are not equal


if USE_CYTHON:
    class TestIStr(IStrMixin):
        cls = istr
